/**
 * @file ptpdebug.c
 * @brief PTP management client program
 * @note Copyright (C) 2012 Richard Cochran <richardcochran@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include <net/if.h>

#include "ds.h"
#include "fsm.h"
#include "notification.h"
#include "pmc_common.h"
#include "print.h"
#include "tlv.h"
#include "uds.h"
#include "util.h"
#include "version.h"
#include "ptp_opt.h"
#include "clock.h"
#include "config.h"
#include "interface.h"
#include "hash.h"
#include "ddt.h"

#define IFMT "\n\t"
#define PTP_SHOW_TIMESTAMP(ts,corr) \
        ((uint64_t)ts.seconds_lsb) | (((uint64_t)ts.seconds_msb) << 32), (ts.nanoseconds + (corr << 16))

static struct pmc *pinfo;

struct hash *htab;

struct interface {
    STAILQ_ENTRY(interface) list;
};

struct config_enum_str {
    const char *label;
    int value;
};

const char *ptp_ps_str[] = {
        "NONE",
        "INIT",
        "FAULTY",
        "DISABLED",
        "LISTEN",
        "PRE_MSTR",
        "MASTER",
        "PASSIVE",
        "UNCALIB",
        "SLAVE",
        "GM",
};

bool is_master_clock = false;
bool debug_enable = false;

static char*
ptp_clock_type_enum2str(int type)
{
    static char buf[64];
    switch (type) {
        case CLOCK_TYPE_ORDINARY:
            snprintf(buf, sizeof(buf), "%s", "Ordinary Clock");
            break;
        case CLOCK_TYPE_BOUNDARY:
            snprintf(buf, sizeof(buf), "%s", "Boundary Clock");
            break;
        case CLOCK_TYPE_P2P:
            snprintf(buf, sizeof(buf), "%s", "Transparent Clock Point-to-Point");
            break;
        case CLOCK_TYPE_E2E:
            snprintf(buf, sizeof(buf), "%s", "Transparent Clock End-to-End");
            break;
    }
    return buf;
}

static char*
ptp_nw_trans_enum2str(int type)
{
    static char buf[64];
    switch (type) {
        case TRANS_UDS:
            snprintf(buf, sizeof(buf), "%s", "Unix Domain Socket");
            break;
        case TRANS_UDP_IPV4:
            snprintf(buf, sizeof(buf), "%s", "IPV4");
            break;
        case TRANS_UDP_IPV6:
            snprintf(buf, sizeof(buf), "%s", "IPV6");
            break;
        case TRANS_IEEE_802_3:
            snprintf(buf, sizeof(buf), "%s", "IEEE-802.3");
            break;
        case TRANS_DEVICENET:
        case TRANS_CONTROLNET:
        case TRANS_PROFINET:
            snprintf(buf, sizeof(buf), "%s(%d)", "Unsupported", type);
            break;
    }
    return buf;
}

static char *
ptp_cid2str(struct ClockIdentity *id)
{
    static char buf[64];
    unsigned char *ptr = id->id;

    snprintf(buf, sizeof(buf),"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
		    ptr[0], ptr[1], ptr[2], ptr[3],
		    ptr[4], ptr[5], ptr[6], ptr[7]);
    return buf;
}

static char *
ptp_pid2str(struct PortIdentity *id)
{
    static char buf[64];
    unsigned char *ptr = id->clockIdentity.id;

    snprintf(buf, sizeof(buf),"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x-%hu",
		    ptr[0], ptr[1], ptr[2], ptr[3],
		    ptr[4], ptr[5], ptr[6], ptr[7],
		    id->portNumber);
    return buf;
}

static char *ptp_text2str(struct PTPText *text)
{
        static struct static_ptp_text s;
        s.max_symbols = -1;
        static_ptp_text_copy(&s, text);
        return (char*)(s.text);
}

const char *ptp_ts_str(enum timestamp_type ts)
{
        switch (ts) {
        case TS_SOFTWARE:
                return "SOFTWARE";
        case TS_HARDWARE:
                return "HARDWARE";
        case TS_LEGACY_HW:
                return "LEGACY_HW";
        case TS_ONESTEP:
                return "ONESTEP";
        case TS_P2P1STEP:
                return "P2P1STEP";
        }

        return "???";
}

static void
debug_arg(PtpParsedArgs *args)
{
    debug_enable = args->debug_enable;
    if (!args->debug_enable)
	    return;
    printf("  CMD Type:%s Iftype:%d\n", 
		    (args->cmd_type == 1) ? "config" :
		    ((args->cmd_type == 2) ? "show" : "clear"), args->interface_type);
    printf("  Global:%s \n", args->global ? "yes" : "no");
    printf("  cfg_opt:%d \n", args->glbl_cfg_opt);
    printf("  show_opt:%d \n", args->glbl_show_opt);
    printf("  command_opt:%d \n", args->glbl_command_opt);
    printf("  Opt Arg: Takes:%d optional:%d present:%d\n",
        args->opt_takes_arg, args->opt_arg_optional,
        args->opt_arg_present);
    printf("  Interface:\n");
    printf("  cfg_opt:%d \n", args->if_cfg_opt);
    printf("  show_opt:%d \n", args->if_show_opt);
    printf("  command_opt:%d \n", args->if_command_opt);

    printf("  Type Arg: %s\n", args->type_argument);
    printf("  Opt STR Arg: %s\n", args->opt_str_argument);
    printf("  Opt BOOL Arg: %d\n", args->opt_bool_argument);
    printf("  Opt INT Arg: %d\n", args->opt_int_argument);
}

static void *ptp_get_mgt_data(struct ptp_message *msg)
{
	struct management_tlv *mgt = (struct management_tlv *) msg->management.suffix;
	return mgt->data;
}

static int ptp_get_mgt_id(struct ptp_message *msg)
{
    struct management_tlv *mgt = (struct management_tlv *) msg->management.suffix;
    return mgt->id;
}

static int ptp_get_mgt_err_id(struct ptp_message *msg)
{
    struct management_error_status *mgt;

    mgt = (struct management_error_status *)msg->management.suffix;
    return mgt->id;
}

static int ptp_is_msg_mgt(struct ptp_message *msg)
{
        struct TLV *tlv;

        if (msg_type(msg) != MANAGEMENT) {
		fprintf(stderr, "Not MANAGEMENT\n");
                return 0;
	}
        if (management_action(msg) != RESPONSE) {
		fprintf(stderr, "Not RESPONSE\n");
                return 0;
	}
        if (msg_tlv_count(msg) != 1) {
		fprintf(stderr, "TLV Count not 1\n");
                return 0;
	}

        tlv = (struct TLV *) msg->management.suffix;
        if (tlv->type == TLV_MANAGEMENT) {
                return 1;
	}
        if (tlv->type == TLV_MANAGEMENT_ERROR_STATUS) {
		fprintf(stderr, "TLV_MANAGEMENT_ERROR_STATUS\n");
                return -1;
	}
        return 0;
}

static int
ptp_send_command(int command, int code, void *data, int data_sz, 
				struct ptp_message **msg)
{
#define N_FD 2
    struct pollfd pollfd[N_FD];
    int  cnt, tmo = 1000, res, loop_cnt = 0;
    bool command_sent = false;
    
    if (command == SET) {
	   if  (!data)  { 
		   pr_err("data should not be null for set\n"); 
		   return -1; 
	   } 
	   pmc_send_set_action(pinfo, code, data, data_sz);
	   return 0;
    } else if (command == COMMAND) {
	    pmc_send_command_action(pinfo, code); 
	   return 0;
    }

    while (1) { 
	    pollfd[0].fd = pmc_get_transport_fd(pinfo); 
	    pollfd[0].events = POLLIN | POLLPRI; 
	    if (!command_sent) 
		    pollfd[0].events |= POLLOUT; 
	    cnt = poll(pollfd, N_FD, tmo); 
	    if (cnt <= 0) { 
		    fprintf(stderr, "Poll cnt <= 0\n");
		    return -1; 
	    } 
	    
	    if (loop_cnt++ > 50) {
		    fprintf(stderr, "Command %s failed\n",
				    command_sent ? "RESPONSE" : "SEND");
		    return -1;
            }

            if (debug_enable)
	        fprintf(stdout, "%d Poll: 0x%x\n", loop_cnt, pollfd[0].revents);
	    if (pollfd[0].revents & POLLOUT) {
		    pmc_send_get_action(pinfo, code); 
		    if (debug_enable)
		    fprintf(stdout, "Sent GET: 0x%x\n", code);
		    command_sent = true; 
	    } 
	    
	    if (pollfd[0].revents & (POLLIN|POLLPRI)) {
	        *msg = pmc_recv(pinfo); 
	        res = ptp_is_msg_mgt(*msg); 
	        if (res < 0 && ptp_get_mgt_err_id(*msg) == code) { 
		    fprintf(stderr, "res:%d code: %d\n", res, code);
		    return -1; 
	        } 
	        if (res <= 0 || ptp_get_mgt_id(*msg) != code) { 
		    fprintf(stderr, "res:%d not code: %d\n", res, code);
		    msg_put(*msg); 
		    *msg = NULL; 
		    continue; 
		}
		break;
	    } 
    }
    return 0;
}


static int ptp_get_number_ports()
{
        struct ptp_message *msg;
        int res;
        struct defaultDS *dds;

        if (debug_enable)
	    fprintf(stdout, "%s:\n", __FUNCTION__); 

        res = ptp_send_command(GET, TLV_DEFAULT_DATA_SET, NULL, 0, &msg);
        if (res < 0)
                return res;

        dds = (struct defaultDS *)ptp_get_mgt_data(msg);
        res = dds->numberPorts;

        msg_put(msg);
        if (debug_enable)
	    fprintf(stdout, "%s: Number of ports:%d\n", __FUNCTION__, res);
        return res;
}

static int ptp_get_port_properties()
{
	struct port_properties_np *ppn, *ppnp; 
	int num_ports, res, i; 
	struct ptp_message *msg; 
	char iface[256];

        if (debug_enable)
	    fprintf(stdout, "%s:\n", __FUNCTION__); 

	num_ports = ptp_get_number_ports();
	if (num_ports <= 0) {
		pr_err("failed to get number of ports");
		return -1;
	}

	for (i = 1; i <= num_ports; i++) { 
		pmc_target_port(pinfo, i); 
		while (1) { 
			res = ptp_send_command(GET, TLV_PORT_PROPERTIES_NP, NULL, 0, &msg); 
			if (res < 0) 
				goto out; 
			ppnp = (struct port_properties_np *)ptp_get_mgt_data(msg); 
			if (ppnp->portIdentity.portNumber != i) { 
				msg_put(msg); 
				continue; 
			} 
			ppn = calloc(1, sizeof(struct port_properties_np) + 
							ppnp->interface.length); 
			if (!ppn) { 
				fprintf(stderr, "low memory, db build failed\n"); 
				return -1; 
			} 
			memcpy(ppn, ppnp, sizeof(struct port_properties_np) + 
						ppnp->interface.length); 
                        if (debug_enable)
			    fprintf(stdout, "%s: portproperty[%d]: %s length: %d\n", 
				__FUNCTION__, i, ptp_text2str(&ppn->interface), 
				ppnp->interface.length); 
			snprintf(iface, sizeof(iface), "%s", 
						ptp_text2str(&ppn->interface)); 
			if (hash_insert(htab, iface, ppn)) { 
				fprintf(stderr, 
					"%s: low memory or duplicate[%d] %s\n", 
					__FUNCTION__, i, iface); 
				free(ppn); 
				return -1; 
			} 
            if (debug_enable)
			    fprintf(stdout, "%s: Inserted port[%d]: %s\n", __FUNCTION__, 
					i, iface); 
			msg_put(msg); 
			res = 1; 
			break; 
		}
	}
out:
        pmc_target_all(pinfo);
        return res;
}

static int
ptp_glbl_show(struct config *cfg, PtpParsedArgs *args, FILE *fp)
{ 
	struct ptp_message *msg1 = NULL, *msg2 = NULL; 
	struct acceptable_master_address_tbl_tlv *amt; 
	char   addr_buf[INET6_ADDRSTRLEN] = {};
	struct acceptableMasterPortDS *amp;
	struct clock_violation_np* vio;
	struct port_properties_np *ppn; 
	struct timePropertiesDS *tp; 
	struct management_tlv *mgt; 
	struct port_stats_np *psp;
	struct interface *iface; 
	struct defaultDS *dds; 
	struct currentDS *cds; 
	struct parentDS *pds; 
	char   date_buf[64];
    char   buf[32];
	struct tm ntm;
	uint16_t val, len; 
	int i, res; 
	uint8_t *p;
    //struct tlv_extra *extra;

	switch (args->glbl_show_opt) { 
		case PTP_ARG_GLBL_SHOW_CLOCK: 
			if (args->opt_arg_present) { 
				res = ptp_send_command(GET, 
					TLV_PARENT_DATA_SET, NULL, 0, &msg1);
                if (!msg1 || res) {
                    fprintf(stderr, "command received no response\n");
                    return -1;
                } 
				
				pds = (struct parentDS *)ptp_get_mgt_data(msg1);
        	    
                fprintf(fp, "Parent Clock"
                    IFMT "Parent Port identity             : %s"
                    IFMT "Parent Offset Scaled Log Variance: 0x%04hx"
                    IFMT "Parent Clock Phase Change Rate   : 0x%08x"
                    IFMT "Parent Stats                     : %hhu"
                    IFMT "GM Identity                      : %s"
                    IFMT "GM Priority1                     : %hhu"
                    IFMT "GM Priority2                     : %hhu"
                    IFMT "GM Clock Quality"                         
                    IFMT "    Class                        : %hhu"
                    IFMT "    Accuracy                     : %hhu"
                    IFMT "GM Offset(log variance)          : 0x%02hhx\n\n",
		    		ptp_pid2str(&pds->parentPortIdentity),
				    pds->observedParentOffsetScaledLogVariance,
				    pds->observedParentClockPhaseChangeRate,
					pds->parentStats, ptp_cid2str(&pds->grandmasterIdentity),
				    pds->grandmasterPriority1, pds->grandmasterPriority2,
				    pds->grandmasterClockQuality.clockClass,
				    pds->grandmasterClockQuality.clockAccuracy, 
					pds->grandmasterClockQuality.offsetScaledLogVariance);
					msg_put(msg1);
            } else { 
			    res = ptp_send_command(GET, TLV_DEFAULT_DATA_SET, 
					    NULL, 0, &msg1); 
			    if (!msg1 || res) { 
				    fprintf(stderr, "command received no response\n"); 
				    return -1; 
			    } 
			    dds = (struct defaultDS *)ptp_get_mgt_data(msg1); 
			    res = ptp_send_command(GET, TLV_CURRENT_DATA_SET, 
					    NULL, 0, &msg2); 
			    cds = (struct currentDS *)ptp_get_mgt_data(msg2); 
		    
			    fprintf(fp, "Clock Information\n" 
				    IFMT "Clock Identity       : %s" 
				    IFMT "Clock Domain         : %d" 
				    IFMT "One Step/Two Step    : %s" 
				    IFMT "Number of PTP ports  : %hu" 
				    IFMT "Priority1            : %hhu" 
				    IFMT "Priority2            : %hhu" 
				    IFMT "Clock Quality" 
				    IFMT "    Class            : %hhu" 
				    IFMT "    Accuracy         : 0x%02hhx" 
				    IFMT "Offset (log variance): 0x%04hx" 
				    IFMT "Offset From Master   : %.1f ns" 
				    IFMT "Mean Path Delay      : %.1f ns" 
				    IFMT "Steps Removed        : %hd\n\n",
                    ptp_cid2str(&dds->clockIdentity),
                    dds->domainNumber,
                    dds->flags & DDS_TWO_STEP_FLAG ? "Two Step" : "One Step",
                    dds->numberPorts,
                    dds->priority1,
                    dds->priority2,
                    dds->clockQuality.clockClass,
                    dds->clockQuality.clockAccuracy,
                    dds->clockQuality.offsetScaledLogVariance,
                    cds->offsetFromMaster/65536.0,
                    cds->meanPathDelay / 65536.0,
                    cds->stepsRemoved);
        	    msg_put(msg1); 
			    msg_put(msg2);
	    	}
            break;
        case PTP_ARG_GLBL_SHOW_CONFIG: 
		    res = ptp_send_command(GET, TLV_DEFAULT_DATA_SET, 
				    NULL, 0, &msg1);
   	         if (!msg1) {
                fprintf(stderr, "command received no response\n");
                return -1;
   	         } 
			 
			 dds = (struct defaultDS *)ptp_get_mgt_data(msg1); 
			 
			 fprintf(fp,
                "PTP mode                : %s\n"
                "Domain                  : %d\n"
                "Clock identity          : %s\n"
                "UTC Offset              : %d\n"
                "----------------------------------------------------------------------------------\n"
                "Port        Po     VLAN    VRF    Transport MsgMode   AMT      State\n"
                "----------------------------------------------------------------------------------\n",
                 ptp_clock_type_enum2str(config_get_int(cfg, NULL, "clock_type")),
                 dds->domainNumber,
                 ptp_cid2str(&dds->clockIdentity),
                 config_get_int(cfg, NULL, "utc_offset"));
			STAILQ_FOREACH(iface, &cfg->interfaces, list) {
			ppn = hash_lookup(htab, interface_netdev(iface)); 
			if (!ppn) { 
				fprintf(fp, "%s: %s not found in hash\n", __FUNCTION__,
						interface_name(iface)); 
				continue; 
			} 
			pmc_target(pinfo, &ppn->portIdentity);
			res = ptp_send_command(GET, TLV_ACCEPTABLE_MASTER_TABLE_ENABLED,
						NULL, 0, &msg2);
			amp = (struct acceptableMasterPortDS *)ptp_get_mgt_data(msg2);
			pmc_target_all(pinfo);
			msg_put(msg2);
			fprintf(fp, "%-11s %-6s %-7s %-6s %-10s %-9s %-8s %-8s(%s)\n", 
		    	interface_name(iface), "N/A", "N/A", "N/A", 
		    	ptp_nw_trans_enum2str(config_get_int(cfg, 
				    interface_name(iface), "network_transport")), 
				config_get_int(cfg, 
					interface_name(iface), "unicast_listen") ? "Unicast" : 
				(config_get_int(cfg, interface_name(iface), 
			    	"hybrid_e2e") ? "Mixed" : "Multicast"), 
				amp->enable ? "Enabled" : "Disabled",
				ptp_ps_str[ppn->port_state],
		    	config_get_int(cfg, interface_name(iface), "masterOnly") ?  
		    		"Master Only" : config_get_int(cfg, NULL, "slaveOnly") ?  
					"Slave Only" : "BMCA"); 
				ppn = NULL;
            }
	    	msg_put(msg1);
            break;
        case PTP_ARG_GLBL_SHOW_STATUS:
            fprintf(fp,
                "----------------------------------------------------\n"
                "Port                 State\n"
                "----------------------------------------------------\n");
			STAILQ_FOREACH(iface, &cfg->interfaces, list) {
				ppn = hash_lookup(htab, interface_netdev(iface)); 
				if (!ppn) { 
					fprintf(fp, "%s: %s not found in hash\n", __FUNCTION__,
							interface_name(iface)); 
					continue; 
				} 
				fprintf(fp, "%-20s %-9s \n", 
			    	interface_name(iface),  ptp_ps_str[ppn->port_state]);
				ppn = NULL;
			}
			break;
        case PTP_ARG_GLBL_SHOW_COUNTERS:
		    fprintf(fp, 
			    "-------------------------------------------------------------------------------\n" 
			    "Packet Type                          Received             Transmitted\n" 
			    "-------------------------------------------------------------------------------\n");
		    STAILQ_FOREACH(iface, &cfg->interfaces, list) {
				ppn = hash_lookup(htab, interface_netdev(iface)); 
				if (!ppn) { 
				    fprintf(fp, "%s: %s not found in hash\n", __FUNCTION__,
						interface_name(iface)); 
				    continue; 
				} 
		        pmc_target(pinfo, &ppn->portIdentity);
				res = ptp_send_command(GET, TLV_PORT_STATS_NP, 
						NULL, 0, &msg1); 
				if (!msg1) { 
					fprintf(stderr, "command received no response\n"); 
					pmc_target_all(pinfo);
					return -1; 
				} 
				mgt = (struct management_tlv *) msg1->management.suffix;
		        psp = (struct port_stats_np *)mgt->data; 
				
				fprintf(fp, "Port        %s\n"
					"  %-30s %20" PRIu64 "%20" PRIu64
					"\n  %-30s %20" PRIu64 "%20" PRIu64
					"\n  %-30s %20" PRIu64 "%20" PRIu64 
					"\n  %-30s %20" PRIu64 "%20" PRIu64 
					"\n  %-30s %20" PRIu64 "%20" PRIu64 
					"\n  %-30s %20" PRIu64 "%20" PRIu64 
					"\n  %-30s %20" PRIu64 "%20" PRIu64 
					"\n  %-30s %20" PRIu64 "%20" PRIu64 
					"\n  %-30s %20" PRIu64 "%20" PRIu64"\n\n", 
					interface_name(iface),
					"Announce", psp->stats.rxMsgType[ANNOUNCE],
				    psp->stats.txMsgType[ANNOUNCE],
					"Sync", psp->stats.rxMsgType[SYNC],
			   		psp->stats.txMsgType[SYNC],	
					"Follow-up", psp->stats.rxMsgType[FOLLOW_UP], 
					psp->stats.txMsgType[FOLLOW_UP],
					"Delay Request", psp->stats.rxMsgType[DELAY_REQ],
			   		psp->stats.txMsgType[DELAY_REQ],	
					"Delay Response", psp->stats.rxMsgType[DELAY_RESP],
			   		psp->stats.txMsgType[DELAY_RESP],	
					"Peer Delay Request", psp->stats.rxMsgType[PDELAY_REQ],
			   		psp->stats.txMsgType[PDELAY_REQ],	
					"Peer Delay Response", psp->stats.rxMsgType[PDELAY_RESP],
					psp->stats.txMsgType[PDELAY_RESP],
					"Management", psp->stats.rxMsgType[MANAGEMENT],
			   		psp->stats.txMsgType[MANAGEMENT],	
					"Signaling", psp->stats.rxMsgType[SIGNALING],
			   		psp->stats.txMsgType[SIGNALING]);	
		    }
		    pmc_target_all(pinfo);
		    break;
        case PTP_ARG_GLBL_SHOW_ACC_MSTR: 
			if (args->opt_arg_present) { 
				res = ptp_send_command(GET, 
								TLV_ACCEPTABLE_MASTER_LOG_NP, 
								NULL, 0, &msg1);
   	         	if (!msg1) {
   	            	fprintf(stderr, "command received no response\n");
   	            	return -1;
   	         	}
			    len = 0;
		    	fprintf(fp, 
					"Acceptable Masters Violation Log\n"
			    	"----------------------------------------------------------------------------------------------------------------\n" 
					"Port             VLAN Clock-ID                 Address                                  Count    Last Occurrence\n" 
			    	"---------------- ---- ------------------------ ---------------------------------------- -------- ---------------\n");
   	         	mgt = (struct management_tlv *) msg1->management.suffix;
				p = (uint8_t *)mgt->data;
				len = mgt->length - 2;
				while (len) {
					vio = (struct clock_violation_np *)p;
					localtime_r(&vio->last_time.tv_sec, &ntm);
					strftime(buf, 30, "%Y %m:%d %H:%M:%S", &ntm);
                    sprintf(date_buf, "%19s.%06d", buf,
                        (int)vio->last_time.tv_usec);

					if (vio->sa_family == AF_INET) {
		                inet_ntop(AF_INET, &(vio->sin_addr), addr_buf, sizeof(addr_buf));
			        } else if (vio->sa_family == AF_INET6) {
		                inet_ntop(AF_INET6, &(vio->sin6_addr), addr_buf, sizeof(addr_buf));
			        } else if (vio->sa_family == AF_PACKET) {
			            snprintf(addr_buf, sizeof(addr_buf), 
								"%02X:%02X:%02X:%02X:%02X:%02X",
			                    vio->sll_addr[0],
			                    vio->sll_addr[1],
			                    vio->sll_addr[2],
			                    vio->sll_addr[3],
			                    vio->sll_addr[4],
			                    vio->sll_addr[5]);
			        }
					fprintf(fp, "%-16s %-4d %-24s %-40s %8d %-20s\n",
						vio->violation_ifname,
						vio->vlan,
						cid2str(&vio->violating_clockId),
						addr_buf,
						vio->count,
						date_buf);
					len -= sizeof(struct clock_violation_np);
					p += sizeof(struct clock_violation_np);
                }
				break;
			}

		    res = ptp_send_command(GET, 
				    TLV_ACCEPTABLE_MASTER_MAX_TABLE_SIZE, 
				    NULL, 0, &msg1);
   	         if (!msg1) {
   	             fprintf(stderr, "command received no response\n");
   	             return -1;
   	         }
   	         mgt = (struct management_tlv *) msg1->management.suffix;
   	         if (mgt->id != TLV_ACCEPTABLE_MASTER_MAX_TABLE_SIZE) {
   	             fprintf(stderr, 
				"mgmt id not TLV_ACCEPTABLE_MASTER_MAX_TABLE_SIZE\n");
   	             return -1;
   	         }
   	         memcpy(&val, mgt->data, sizeof(uint16_t));
   	         fprintf(fp, "Acceptable Master Table Max Size: %d\n", val);
	
				res = ptp_send_command(GET, TLV_ACCEPTABLE_MASTER_TABLE, 
										NULL, 0, &msg2);
   	         if (!msg2) {
   	             fprintf(stderr, "command received no response\n");
   	             return -1;
   	         } 
			 
			 amt = (struct acceptable_master_address_tbl_tlv *)ptp_get_mgt_data(msg2); 
			 fprintf(fp, "Num Clock ID                Alternate Priority1\n"
                        "--- ----------------------- ------------------\n"); 
			 for (i = 0; i < amt->tableSize; i++) { 
				fprintf(fp, "%3d %s %3d\n", 
					i+1, ptp_cid2str(&amt->masterTable[i].clockIdentity),
                    amt->masterTable[i].alternatePriority1); 
			 } 
			 msg_put(msg1); 
			 msg_put(msg2); 
			 break;
        case PTP_ARG_GLBL_SHOW_TIME_PROPERTY: 
			 
			 res = ptp_send_command(GET, TLV_TIME_PROPERTIES_DATA_SET, 
									NULL, 0, &msg1); 
			 if (!msg1) { 
				fprintf(stderr, "command received no response\n");
                return -1; 
			 } 
			 tp = (struct timePropertiesDS *)ptp_get_mgt_data(msg1); 
			 fprintf(fp, "Time Properties "
                    IFMT "Current UTC Offset Valid : %d (%s)"
                    IFMT "Current UTC Offset       : %hd"
                    IFMT "Leap61                   : %d (%s)"
                    IFMT "Leap59                   : %d (%s)"
                    IFMT "Time Traceable           : %d (%s)"
                    IFMT "Frequency Traceable      : %d (%s)"
                    IFMT "PTP Timescale            : %d (%s)"
                    IFMT "Time Source              : 0x%02hhx\n\n",
                    tp->flags & UTC_OFF_VALID ? 1 : 0,
                    tp->flags & UTC_OFF_VALID ? "True" : "False",
                    tp->currentUtcOffset,
                    tp->flags & LEAP_61 ? 1 : 0,
                    tp->flags & LEAP_61 ? "True" : "False",
                    tp->flags & LEAP_59 ? 1 : 0,
                    tp->flags & LEAP_59 ? "True" : "False",
                    tp->flags & TIME_TRACEABLE ? 1 : 0,
                    tp->flags & TIME_TRACEABLE ? "True" : "False",
                    tp->flags & FREQ_TRACEABLE ? 1 : 0,
                    tp->flags & FREQ_TRACEABLE ? "True" : "False",
                    tp->flags & PTP_TIMESCALE ? 1 : 0,
                    tp->flags & PTP_TIMESCALE ? "True" : "False",
                    tp->timeSource); 
		    msg_put(msg1);
            break;
		case PTP_ARG_GLBL_SHOW_FORCED_MSTR:

			if (args->opt_arg_present) {
				res = ptp_send_command(GET, 
							TLV_FORCED_MASTER_LOG_NP,
								NULL, 0, &msg1);
   	         	if (!msg1) {
   	            	fprintf(stderr, "command received no response\n");
   	            	return -1;
   	         	}
			    len = 0;
		    	fprintf(fp, 
					"Forced Master Violation Log\n"
			    	"----------------------------------------------------------------------------------------------------------------\n" 
					"Port             VLAN Clock-ID                 Address                                  Count    Last Occurrence\n" 
			    	"---------------- ---- ------------------------ ---------------------------------------- -------- ---------------\n");
   	         	mgt = (struct management_tlv *) msg1->management.suffix;
				p = (uint8_t *)mgt->data;
				len = mgt->length - 2;
				while (len) {
					vio = (struct clock_violation_np *)p;
					localtime_r(&vio->last_time.tv_sec, &ntm);
					strftime(buf, 32, "%Y %m:%d %H:%M:%S", &ntm);
                    sprintf(date_buf,"%19s.%06d", buf,
                        (int)vio->last_time.tv_usec);
					if (vio->sa_family == AF_INET) {
		                inet_ntop(AF_INET, &(vio->sin_addr), addr_buf, sizeof(addr_buf));
			        } else if (vio->sa_family == AF_INET6) {
		                inet_ntop(AF_INET6, &(vio->sin6_addr), addr_buf, sizeof(addr_buf));
			        } else if (vio->sa_family == AF_PACKET) {
			            snprintf(addr_buf, sizeof(addr_buf), 
								"%02X:%02X:%02X:%02X:%02X:%02X",
			                    vio->sll_addr[0],
			                    vio->sll_addr[1],
			                    vio->sll_addr[2],
			                    vio->sll_addr[3],
			                    vio->sll_addr[4],
			                    vio->sll_addr[5]);
			        }
					fprintf(fp, "%-16s %-4d %-24s %-40s %8d %-32s\n",
						vio->violation_ifname,
						vio->vlan,
						cid2str(&vio->violating_clockId),
						addr_buf,
						vio->count,
						date_buf);
					len -= sizeof(struct clock_violation_np);
					p += sizeof(struct clock_violation_np);
                }
				break;
			}

		    STAILQ_FOREACH(iface, &cfg->interfaces, list) {
				if (config_get_int(cfg, interface_name(iface), "masterOnly")) {
  				    fprintf(fp, "%-11s\n", interface_name(iface));
				}
			}
			break;
		default:
            break;
    }

    return 0;
}

static int
ptp_if_show(struct config *cfg, PtpParsedArgs *args, FILE *fp)
{ 
	struct ptp_message *msg1 = NULL, *msg2 = NULL;
	struct port_properties_np *pp_np;
	struct port_properties_np *ppn;
	struct foreign_master *fm;
	struct port_stats_np *psp;
	struct port_ds_np *pds_np;
	struct portDS *pds; 
	int res;

        if (args->debug_enable)
            fprintf(stdout, "%s:\n", __FUNCTION__); 

	ppn = hash_lookup(htab, args->type_argument); 
	if (!ppn) { 
		fprintf(fp, "%s: %s not found in hash\n", __FUNCTION__, args->type_argument); 
		return -1; 
	} 
	pmc_target(pinfo, &ppn->portIdentity);
	switch (args->if_show_opt) { 
		case PTP_ARG_IF_SHOW_CONFIG:
#if 0
Interface name:                         Eth1/4
		Channel group ID:                       N/A
		VRF:                                    default
		IP Address:                             4.4.4.4/24
		PTP Port number:                        1
		Forced Master:                          no
		Transport protocol:                     UDP IPv4
		IPv6 Multicast scope ID:                N/A
#endif
			res = ptp_send_command(GET, TLV_PORT_DATA_SET, 
					NULL, 0, &msg1); 
			if (!msg1 || res) { 
				fprintf(stderr, "command received no response\n"); 
				pmc_target_all(pinfo);
				return -1; 
			} 
			pds = (struct portDS *)ptp_get_mgt_data(msg1);

			fprintf(fp, 
		        "Port identity:                          %s\n"
		        "Announce interval(log mean):            %hhd\n"
		        "Sync interval(log mean):                %hhd\n" 
		        "Delay request interval(log mean):       %hhd\n"
		        "Peer Delay request interval(log mean):  %hhd\n" 
				"Delay Mechanism:                        %s\n"
		        "Announce receipt time out:              %hhu\n",
				ptp_pid2str(&pds->portIdentity), pds->logAnnounceInterval,
				pds->logSyncInterval, pds->logMinDelayReqInterval,
				pds->logMinPdelayReqInterval, 
				(pds->delayMechanism == DM_AUTO) ? "Auto" :
				(pds->delayMechanism == DM_E2E) ? "End to End" : "Point to Point",
				pds->announceReceiptTimeout);
			msg_put(msg1);
			break;
		case PTP_ARG_IF_SHOW_STATUS:
			res = ptp_send_command(GET, TLV_PORT_DATA_SET_NP, 
					NULL, 0, &msg1); 
			if (!msg1) { 
				fprintf(stderr, 
					"command received no response\n"); 
				pmc_target_all(pinfo);
				return -1; 
			} 
			pds_np = (struct port_ds_np *)ptp_get_mgt_data(msg1); 
			res = ptp_send_command(GET, TLV_PORT_PROPERTIES_NP, 
				NULL, 0, &msg2); 
			if (!msg2) { 
				fprintf(stderr, "command received no response\n"); 
				pmc_target_all(pinfo);
				return -1; 
			} 
			pp_np = (struct port_properties_np *)ptp_get_mgt_data(msg2);
			fprintf(fp, 
		        "Port identity:                                 %s\n"
		        "State:                                         %s\n"
		        "Time stamping:                                 %s\n"
		        "IEEE 802.1AS capable:                          %s\n"
		        "Neighbor Propagation Delay Threshold stamping: %d\n",
				ptp_pid2str(&pp_np->portIdentity), 
				ptp_ps_str[pp_np->port_state],
				ptp_ts_str(pp_np->timestamping),
				pds_np->asCapable ? "Yes" : "No",
				pds_np->neighborPropDelayThresh);
			msg_put(msg1);
			msg_put(msg2);
			break;

		case PTP_ARG_IF_SHOW_COUNTERS:
			res = ptp_send_command(GET, TLV_PORT_STATS_NP, 
					NULL, 0, &msg1); 
			if (!msg1) { 
				fprintf(stderr, "command received no response\n"); 
				pmc_target_all(pinfo);
				return -1; 
			} 
			psp = (struct port_stats_np *)ptp_get_mgt_data(msg1);
			fprintf(fp, 
				    "-------------------------------------------------------------------------------\n" 
				    "Packet Type                          Received             Transmitted\n" 
				    "-------------------------------------------------------------------------------\n"
					"%-30s %20" PRIu64 "%20" PRIu64
					"\n%-30s %20" PRIu64 "%20" PRIu64
					"\n%-30s %20" PRIu64 "%20" PRIu64 
					"\n%-30s %20" PRIu64 "%20" PRIu64 
					"\n%-30s %20" PRIu64 "%20" PRIu64 
					"\n%-30s %20" PRIu64 "%20" PRIu64 
					"\n%-30s %20" PRIu64 "%20" PRIu64 
					"\n%-30s %20" PRIu64 "%20" PRIu64 
					"\n%-30s %20" PRIu64 "%20" PRIu64"\n", 
					"Announce", psp->stats.rxMsgType[ANNOUNCE],
					    psp->stats.txMsgType[ANNOUNCE],
					"Sync", psp->stats.rxMsgType[SYNC],
				   		psp->stats.txMsgType[SYNC],	
					"Follow-up", psp->stats.rxMsgType[FOLLOW_UP], 
						psp->stats.txMsgType[FOLLOW_UP],
					"Delay Request", psp->stats.rxMsgType[DELAY_REQ],
				   		psp->stats.txMsgType[DELAY_REQ],	
					"Delay Response", psp->stats.rxMsgType[DELAY_RESP],
				   		psp->stats.txMsgType[DELAY_RESP],	
					"Peer Delay Request", psp->stats.rxMsgType[PDELAY_REQ],
				   		psp->stats.txMsgType[PDELAY_REQ],	
					"Peer Delay Response", psp->stats.rxMsgType[PDELAY_RESP],
						psp->stats.txMsgType[PDELAY_RESP],
					"Management", psp->stats.rxMsgType[MANAGEMENT],
				   		psp->stats.txMsgType[MANAGEMENT],	
					"Signaling", psp->stats.rxMsgType[SIGNALING],
				   		psp->stats.txMsgType[SIGNALING]);	
			break;
		case PTP_ARG_IF_SHOW_FOREIGN_MSTRS:
			res = ptp_send_command(GET, TLV_FOREIGN_MASTERS_DATA_SET, 
					NULL, 0, &msg1); 
			if (!msg1) { 
				fprintf(stderr, "command received no response\n"); 
				pmc_target_all(pinfo);
				return -1; 
			} 
			fm = (struct foreign_master *)ptp_get_mgt_data(msg1);
			fprintf(fp, "FOREIGN_MASTER "
                            IFMT "clockIdentity           %s"
                            IFMT "priority1               %u"
                            IFMT "priority2               %u"
                            IFMT "clockClass              %hhu"
                            IFMT "clockAccuracy           0x%02hhx"
                            IFMT "offsetScaledLogVariance 0x%04hx"
                            IFMT "stepsRemoved            %hd"
                            IFMT "isGmc                   %u"
                            IFMT "linux_ifindex           %u",
                            cid2str(&fm->identity),
                            fm->priority1,
                            fm->priority2,
                            fm->quality.clockClass,
                            fm->quality.clockAccuracy,
                            fm->quality.offsetScaledLogVariance,
                            fm->stepsRemoved,
                            fm->isGmc,
                            fm->linux_ifindex);
			break;
	}
	pmc_target_all(pinfo);
        if (args->debug_enable)
            fprintf(stdout, "%s: Done\n", __FUNCTION__); 
	return 0;
}

static int
ptp_glbl_cfg(struct config *cfg, PtpParsedArgs *args, FILE *fp)
{ 
	struct management_tlv_datum mtd;
	struct ptp_message *msg = NULL;

	switch (args->glbl_cfg_opt) {
        case PTP_ARG_GLBL_CFG_PTP:
		break;
        case PTP_ARG_GLBL_CFG_DOMAIN:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_DOMAIN, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_GLBL_CFG_PRIO1:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_PRIORITY1, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_GLBL_CFG_PRIO2:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_PRIORITY2, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_GLBL_CFG_OFF_MIN_THRSH:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_OFFSET_MIN_THRESHOLD_NP, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_GLBL_CFG_OFF_MAX_THRSH:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_OFFSET_MAX_THRESHOLD_NP, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_GLBL_CFG_MPD_THRSH:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_MPD_THRESHOLD_NP, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_GLBL_CFG_ACC_MSTR:
	case PTP_ARG_GLBL_CFG_MSG_MODE: 
	case PTP_ARG_GLBL_CFG_SLAVE_ONLY: 
	case PTP_ARG_GLBL_CFG_TTL: 
	case PTP_ARG_GLBL_CFG_DSCP_EVENT: 
	case PTP_ARG_GLBL_CFG_DSCP_GEN: 
	default:
		break;
	}

	return 0;
}

static int
ptp_if_cfg(struct config *cfg, PtpParsedArgs *args, FILE *fp)
{ 
	struct management_tlv_datum mtd;
	struct port_properties_np *ppn;
	struct ptp_message *msg = NULL;

        if (args->debug_enable)
            fprintf(stdout, "%s:\n", __FUNCTION__); 

	ppn = hash_lookup(htab, args->type_argument); 
	if (!ppn) { 
		fprintf(fp, "%s: %s not found in hash\n", __FUNCTION__, args->type_argument); 
		return -1; 
	} 
	pmc_target(pinfo, &ppn->portIdentity);
	switch (args->if_cfg_opt) {
	case PTP_ARG_IF_CFG_PTP:
	case PTP_ARG_IF_CFG_IPV6:
	case PTP_ARG_IF_CFG_FORCED_MSTR:
		mtd.val = args->opt_bool_argument;
		ptp_send_command(SET, TLV_MASTER_ONLY, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_IF_CFG_ANN_INTVL:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_LOG_ANNOUNCE_INTERVAL, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_IF_CFG_SYNC_INTVL:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_LOG_SYNC_INTERVAL, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_IF_CFG_DELAY_REQ_INTVL:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_LOG_MIN_PDELAY_REQ_INTERVAL, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_IF_CFG_ANN_TIMEOUT:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_ANNOUNCE_RECEIPT_TIMEOUT, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_IF_CFG_DELAY_MECHANISM:
		mtd.val = args->opt_int_argument;
		ptp_send_command(SET, TLV_DELAY_MECHANISM, &mtd, sizeof(mtd), &msg);
		break;
	case PTP_ARG_IF_CFG_ACC_MSTR_TBL_ENBL:
		mtd.val = args->opt_bool_argument;
		ptp_send_command(SET, TLV_ACCEPTABLE_MASTER_TABLE_ENABLED, &mtd, sizeof(mtd), &msg);
		break;
	default:
		break;
	}
	pmc_target_all(pinfo);
	return 0;
}

static int
ptp_glbl_command(struct config *cfg, PtpParsedArgs *args, FILE *fp)
{ 
	struct ptp_message *msg = NULL;

        if (args->debug_enable)
            fprintf(stdout, "%s: opt:%d\n", __FUNCTION__, args->glbl_command_opt); 

	switch (args->glbl_command_opt) {
        case PTP_ARG_GLBL_CLEAR_ACC_MSTRS:
		ptp_send_command(COMMAND, TLV_ACCEPTABLE_MASTER_LOG_NP, NULL, 0, &msg);
		break;
        case PTP_ARG_GLBL_CLEAR_FORCED_MSTRS:
		ptp_send_command(COMMAND, TLV_FORCED_MASTER_LOG_NP, NULL, 0, &msg);
		break;
        case PTP_ARG_GLBL_CLEAR_COUNTERS:
		pmc_target_all(pinfo);
		ptp_send_command(COMMAND, TLV_PORT_CLEAR_COUNTERS_NP, NULL, 0, &msg);
		break;
        case PTP_ARG_GLBL_SET_LOGGING:
		if (args->opt_bool_argument) {
		    ptp_send_command(COMMAND, TLV_LOGGING_ENABLE_NP, NULL, 0, &msg);
		} else {
		    ptp_send_command(COMMAND, TLV_LOGGING_DISABLE_NP, NULL, 0, &msg);
		}
		break;
	}
	return 0;
}

static int
ptp_if_command(struct config *cfg, PtpParsedArgs *args, FILE *fp)
{ 
	struct port_properties_np *ppn;
	struct ptp_message *msg = NULL;

        if (args->debug_enable)
            fprintf(stdout, "%s:\n", __FUNCTION__); 

	ppn = hash_lookup(htab, args->type_argument); 
	if (!ppn) { 
		fprintf(fp, "%s: %s not found in hash\n", __FUNCTION__, args->type_argument); 
		return -1; 
	} 
	pmc_target(pinfo, &ppn->portIdentity);
	switch (args->if_command_opt) {
        case PTP_ARG_IF_CLEAR_COUNTERS:
		ptp_send_command(COMMAND, TLV_PORT_CLEAR_COUNTERS_NP, NULL, 0, &msg);
		break;
	}
	pmc_target_all(pinfo);

	return 0;
}

static int
ptp_perform_command(struct config *cfg, PtpParsedArgs *args)
{
    int ret = 0;

    if (args->debug_enable)
        fprintf(stdout, "%s:\n", __FUNCTION__); 

    htab = hash_create(); 
    

    if (args->global) {
        switch(args->cmd_type) {
            case PTP_ARG_CMD_TYPE_SET:
                ret = ptp_glbl_cfg(cfg, args, stdout);
                break;
            case PTP_ARG_CMD_TYPE_GET:
    		ptp_get_port_properties();
                ret = ptp_glbl_show(cfg, args, stdout);
                break;
            case PTP_ARG_CMD_TYPE_COMMAND:
                ret = ptp_glbl_command(cfg, args, stdout);
                break;
        }
    } else {
	ptp_get_port_properties();
        switch(args->cmd_type) {
            case PTP_ARG_CMD_TYPE_SET:
                ret = ptp_if_cfg(cfg, args, stdout);
                break;
            case PTP_ARG_CMD_TYPE_GET:
                ret = ptp_if_show(cfg, args, stdout);
                break;
            case PTP_ARG_CMD_TYPE_COMMAND:
                ret = ptp_if_command(cfg, args, stdout);
                break;
        }
    }
    hash_destroy(htab, NULL);
    if (args->debug_enable)
        fprintf(stdout, "%s: done\n", __FUNCTION__); 
    return ret;
}

int main(int argc, char *argv[])
{
    const char *iface_name = NULL;
    char config[MAX_IFNAME_SIZE + 1];
    int zero_datalen = 0;
    int ret = 0;
    char uds_local[MAX_IFNAME_SIZE + 1];
    struct config *cfg;
    uint8_t domain_number = 0, transport_specific = 0;
    uint8_t boundary_hops = 1;

    PtpParsedArgs args = ptp_cmd_parser(argc, argv);

    debug_arg(&args);

    handle_term_signals();

    cfg = config_create();
    if (!cfg) {
        fprintf(stderr, "failed to create cfg\n");
        return -1;
    }

    snprintf(uds_local, sizeof(uds_local), "/var/run/ptpcli.%d", getpid());
    iface_name = uds_local;

    snprintf(config, sizeof(config), "/etc/ptp4l.conf");
    if (config_read(config, cfg)) {
        fprintf(stderr, "failed to read /etc/ptp4l.conf\n");
        config_destroy(cfg);
        return -1;
    }

    print_set_syslog(1);
    print_set_verbose(1);

    domain_number = config_get_int(cfg, NULL, "domainNumber");

    if (debug_enable)
        fprintf(stdout, "Domain: %d\n", domain_number);

    pinfo = pmc_create(cfg, TRANS_UDS, iface_name, boundary_hops,
             domain_number, transport_specific, zero_datalen);
    if (!pinfo) {
        fprintf(stderr, "failed to create pmc\n");
        config_destroy(cfg);
        return -1;
    }

    ptp_perform_command(cfg, &args); 

    pmc_destroy(pinfo);
    msg_cleanup();

    config_destroy(cfg);
    return ret;
}




