/**
 * @file raw_udp.c
 * @note Copyright (C) 2012 Richard Cochran <richardcochran@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/filter.h>
#include <linux/if_ether.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netpacket/packet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <linux/errqueue.h>
#include <linux/net_tstamp.h>
#include <linux/sockios.h>

#include "address.h"
#include "config.h"
#include "contain.h"
#include "ether.h"
#include "print.h"
#include "raw_udp.h"
#include "raw_udp_common.h"
#include "sk.h"
#include "transport_private.h"
#include "util.h"

#define EVENT_PORT        319
#define GENERAL_PORT      320
#define PTP_PRIMARY_MCAST_IPADDR "224.0.1.129"
#define PTP_PDELAY_MCAST_IPADDR  "224.0.0.107"

struct udp_hdr {
    uint16_t udp_src;
    uint16_t udp_dst;
    uint16_t udp_len;
    uint16_t udp_check;
} __attribute__((packed));

struct ip_hdr {
#if 1
    uint8_t        ip_hl:4,
                   ip_v:4;
#endif
#if 0 /*BIG*/
    uint8_t        ip_v:4,
                   ip_hl:4;
#endif
    uint8_t        ip_tos;
    uint16_t       ip_len;
    uint16_t       ip_id;
    uint16_t       ip_off;
    uint8_t        ip_ttl;
    uint8_t        ip_p;
    uint16_t       ip_check;
    struct in_addr ip_src;
    struct in_addr ip_dst;
} __attribute__((packed));

struct pseudo_hdr {
    uint32_t ip_src;
    uint32_t ip_dst;
    uint8_t  reserved;
    uint8_t  protocol;
    uint16_t length;
} __attribute__((packed));

struct raw {
	struct transport t;
	struct address src_addr;
	struct address ptp_addr;
	struct address p2p_addr;

	struct in_addr src_ip;
	struct address src_mac;

	int      ttl;
	int      use_vlan;
    uint32_t vlan;

    uint8_t dscp_general;
    uint8_t dscp_event;
};

#define OP_AND  (BPF_ALU | BPF_AND | BPF_K)
#define OP_JEQ  (BPF_JMP | BPF_JEQ | BPF_K)
#define OP_JUN  (BPF_JMP | BPF_JA)
#define OP_LDB  (BPF_LD  | BPF_B   | BPF_ABS)
#define OP_LDH  (BPF_LD  | BPF_H   | BPF_ABS)
#define OP_RETK (BPF_RET | BPF_K)

#define PTP_GEN_BIT 0x08 /* indicates general message, if set in message type */

#define N_RAW_UNTAGGED_FILTER          11
#define N_RAW_TAGGED_FILTER            16
#define RAW_PTP_UNTAGGED_MSG_TYPE_TEST 8
#define RAW_PTP_TAGGED_MSG_TYPE_TEST   13
#define RAW_VLAN_TEST         4


/* BPF Code for tagged packets:
  ld vlan_avail
  jneq #0x1, drop_not_ptp
  ld vlan_tci
  and #0x0fff
  jneq #0xa, drop_not_ptp
  ldh [12]                      ; load ethertype
  jneq #0x800, drop_not_ptp     ; ETH_P_IP
  ldb [23]                      ; load proto
  jneq #17, drop_not_ptp        ; IPPROTO_UDP ?
  ldh [20]                      ; load frag offset field
  jset #0x1fff, drop_not_ptp    ; don't allow fragments
  ldxb 4*([14]&0xf)             ; load IP header len
  ldh [x + 16]                  ; load UDP dst port
  jneq #320, drop_not_ptp       ; is port PTP_EV_PORT ?
  ret #-1                       ; return class
  drop_not_ptp: ret #0
*/

/* BPF Code for untagged packets:
  ldh [12]                      ; load ethertype
  jneq #0x800, drop_not_ptp     ; ETH_P_IP
  ldb [23]                      ; load proto
  jneq #17, drop_not_ptp        ; IPPROTO_UDP ?
  ldh [20]                      ; load frag offset field
  jset #0x1fff, drop_not_ptp    ; don't allow fragments
  ldxb 4*([14]&0xf)             ; load IP header len
  ldh [x + 16]                  ; load UDP dst port
  jneq #320, drop_not_ptp       ; is port PTP_EV_PORT ?
  ret #-1                       ; return class
  drop_not_ptp: ret #0
*/

/* Get the BPF
 * ./bpf_asm -c vlan_tci.bpf
 */


static struct sock_filter raw_tagged_filter[N_RAW_TAGGED_FILTER] = {
    { 0x20,  0,  0, 0xfffff030 },
    { 0x15,  0, 13, 0x00000001 },
    { 0x20,  0,  0, 0xfffff02c },
    { 0x54,  0,  0, 0x00000fff },
    { 0x15,  0, 10, 0x0000000a },
    { 0x28,  0,  0, 0x0000000c },
    { 0x15,  0,  8, 0x00000800 },
    { 0x30,  0,  0, 0x00000017 },
    { 0x15,  0,  6, 0x00000011 },
    { 0x28,  0,  0, 0x00000014 },
    { 0x45,  4,  0, 0x00001fff },
    { 0xb1,  0,  0, 0x0000000e },
    { 0x48,  0,  0, 0x00000010 },
    { 0x15,  0,  1, 0x00000140 },
    { 0x06,  0,  0, 0xffffffff },
    { 0x06,  0,  0, 0000000000 },
};

static struct sock_filter raw_untagged_filter[N_RAW_UNTAGGED_FILTER] = {
    { 0x28,  0,  0, 0x0000000c },
    { 0x15,  0,  8, 0x00000800 },
    { 0x30,  0,  0, 0x00000017 },
    { 0x15,  0,  6, 0x00000011 },
    { 0x28,  0,  0, 0x00000014 },
    { 0x45,  4,  0, 0x00001fff },
    { 0xb1,  0,  0, 0x0000000e },
    { 0x48,  0,  0, 0x00000010 },
    { 0x15,  0,  1, 0x00000140 },
    { 0x06,  0,  0, 0xffffffff },
    { 0x06,  0,  0, 0000000000 },
};


static int raw_udp_close(struct transport *t, struct fdarray *fda)
{
	close(fda->fd[0]);
	close(fda->fd[1]);
	if (fda->sock_event > 0) {
		close(fda->sock_event);
	}
	if (fda->sock_general > 0) {
		close(fda->sock_general);
	}
	return 0;
}

static void mac_to_addr(struct address *addr, void *mac)
{
	addr->sll.sll_family = AF_PACKET;
	addr->sll.sll_halen = MAC_LEN;
	memcpy(addr->sll.sll_addr, mac, MAC_LEN);
	addr->len = sizeof(addr->sll);
}

static void addr_to_mac(void *mac, struct address *addr)
{
	memcpy(mac, &addr->sll.sll_addr, MAC_LEN);
}

static void mc_ip_2_mac(struct in_addr ip_addr, unsigned char mac[MAC_LEN])
{
    uint32_t addr = htonl(ip_addr.s_addr);

    mac[0] = 0x1;
    mac[1] = 0x0;
    mac[2] = 0x5E;
    mac[3] = (addr & 0x007F0000) >> 16;
    mac[4] = (addr & 0x0000FF00) >> 8;
    mac[5] = addr & 0x000000FF;
}

enum { MC_PRIMARY, MC_PDELAY };
static struct in_addr mcast_addr[2];


static int raw_udp_open(struct transport *t, struct interface *iface,
        struct fdarray *fda, enum timestamp_type ts_type)
{
	struct raw *raw = container_of(t, struct raw, t);
	unsigned char ptp_dst_mac[MAC_LEN];
	unsigned char p2p_dst_mac[MAC_LEN];
	struct sock_fprog prg;
	//struct address ipaddr;
	int type_test_index;
	int efd, gfd;
	char *str;
	unsigned int vlan = 0;
	char *vid = NULL;
	char netdev[MAX_IFNAME_SIZE + 1] = {0};
	char *vlan_intf = config_get_string(t->cfg, interface_name(iface), "vlan_intf");
	char *lag_intf = config_get_string(t->cfg, interface_name(iface), "lag_intf");
    char *l2_mode = config_get_string(t->cfg, interface_name(iface), "l2_mode");
	char *intf = NULL;
	int sock_event, sock_general;

	if (strlen(vlan_intf)) {
		intf = vlan_intf;
	} else if (strlen(lag_intf)) {
		intf = lag_intf;
	} else {
		intf = "";
	} 
	
	strncpy(netdev, interface_name(iface), MAX_IFNAME_SIZE); 
	vid = strchr(netdev, '@'); 
	if (vid != NULL) { 
		*vid = '\0'; 
	}

	raw->ttl = config_get_int(t->cfg, interface_name(iface), "udp_ttl");
	raw->vlan = vlan = interface_vlan(iface);
	raw->dscp_event = config_get_int(t->cfg, interface_name(iface), "dscp_event");
	raw->dscp_general = config_get_int(t->cfg, interface_name(iface), "dscp_general");

	str = config_get_string(t->cfg, interface_name(iface), "p2p_dst_mac");
	if (str2mac(str, p2p_dst_mac)) {
		pr_err("invalid p2p_dst_mac %s", str);
		return -1;
	} 
	
	if (!inet_aton(PTP_PRIMARY_MCAST_IPADDR, &mcast_addr[MC_PRIMARY])) 
		return -1; 
	
	if (!inet_aton(PTP_PDELAY_MCAST_IPADDR, &mcast_addr[MC_PDELAY])) 
		return -1; 
	
	str = config_get_string(t->cfg, interface_name(iface), "src_ip"); 
	
	if (!inet_aton(str, &raw->src_ip)) 
		return -1;

	mc_ip_2_mac(mcast_addr[MC_PRIMARY], ptp_dst_mac);

	mac_to_addr(&raw->ptp_addr, ptp_dst_mac);
	mac_to_addr(&raw->p2p_addr, p2p_dst_mac);
	if (sk_interface_macaddr(intf, &raw->src_addr))
		goto no_mac;

	raw->use_vlan = 0;
#if 0
	if (sk_interface_addr(interface_name(iface), AF_INET,
	                                &ipaddr) < 1) {
		raw->use_vlan = 1;
	}
#endif
    if (l2_mode &&  (strcmp(l2_mode, "trunk") == 0)) {
		raw->use_vlan = 1;
    }
	if (vlan &&  raw->use_vlan) {
		prg.filter = raw_tagged_filter;
		prg.len = N_RAW_TAGGED_FILTER;
		raw_tagged_filter[RAW_VLAN_TEST].k = vlan;
		type_test_index = RAW_PTP_TAGGED_MSG_TYPE_TEST;
	} else {
		prg.filter = raw_untagged_filter;
		prg.len = N_RAW_UNTAGGED_FILTER;
		type_test_index = RAW_PTP_UNTAGGED_MSG_TYPE_TEST;
	}

	sock_event = raw_udp_common_bind_port_to_socket(EVENT_PORT, 1);
	if (sock_event < 0) {
		pr_warning("failed to bind event port to socket");
	}

	sock_general = raw_udp_common_bind_port_to_socket(GENERAL_PORT, 1);
	if (sock_general < 0) {
		pr_warning("failed to bind general port to socket");
	}

	prg.filter[type_test_index].k = EVENT_PORT;
	efd = raw_udp_common_open_socket(netdev, 1,
					 ptp_dst_mac, p2p_dst_mac, &prg);
	if (efd < 0)
		goto no_event;

	prg.filter[type_test_index].k = GENERAL_PORT;
	gfd = raw_udp_common_open_socket(netdev, 0,
					 ptp_dst_mac, p2p_dst_mac, &prg);
	if (gfd < 0)
		goto no_general;

	if (sk_timestamping_init(efd, netdev, ts_type, TRANS_RAW_UDP_IPV4))
		goto no_timestamping;

	if (sk_general_init(gfd))
		goto no_timestamping;

	fda->fd[FD_EVENT] = efd;
	fda->fd[FD_GENERAL] = gfd;
	fda->sock_event = sock_event;
	fda->sock_general = sock_general;
	return 0;

no_timestamping:
	close(gfd);
no_general:
	close(efd);
no_event:
no_mac:
	return -1;
}

static int raw_udp_recv(struct transport *t, int fd, void *buf, int buflen,
		    struct address *addr, struct hw_timestamp *hwts)
{
	int cnt, hlen;
	unsigned char *ptr = buf;
	struct ip_hdr *ip_hdr = NULL;

    //hlen = sizeof(struct udp_hdr) + sizeof(struct ip_hdr) + sizeof(struct vlan_hdr);
    hlen = sizeof(struct udp_hdr) + sizeof(struct ip_hdr) + sizeof(struct eth_hdr);

	ptr    -= hlen;
	buflen += hlen;

	cnt = sk_receive(fd, ptr, buflen, addr, hwts, 0);

	ip_hdr = (struct ip_hdr *) (ptr + sizeof(struct eth_hdr));
	addr->peer_ip.sin.sin_family = AF_INET;
	addr->peer_ip.sin.sin_addr.s_addr = ip_hdr->ip_src.s_addr;

	if (cnt >= 0)
		cnt -= hlen;
	if (cnt < 0)
		return cnt;
	return cnt;
}

static uint16_t udp_checksum(struct ip_hdr *ip_hdr, void *buf, int udp_length) {
	unsigned char pkt[1600] = {0};
	struct pseudo_hdr ps_hdr = {0};

	ps_hdr.ip_src = ip_hdr->ip_src.s_addr;
	ps_hdr.ip_dst = ip_hdr->ip_dst.s_addr;
	ps_hdr.reserved = 0;
	ps_hdr.protocol = IPPROTO_UDP;
	ps_hdr.length = htons(udp_length);

	memcpy(pkt, &ps_hdr, sizeof(ps_hdr));
	memcpy(pkt + sizeof(ps_hdr), buf, udp_length);

	return raw_udp_common_checksum(pkt, sizeof(ps_hdr) + udp_length);
}

static int raw_udp_send(struct transport *t, struct fdarray *fda, 
	            enum transport_event event, int peer, void *buf, 
		    int len, struct address *addr, struct hw_timestamp *hwts)
{
	struct raw *raw = container_of(t, struct raw, t);
	struct ptp_message *msg = buf;
	ssize_t cnt;
	int fd = event ? fda->fd[FD_EVENT] : fda->fd[FD_GENERAL];
	unsigned char pkt[1600], *ptr = buf;
	struct vlan_hdr *vlan_hdr;
	struct eth_hdr  *eth_hdr;
	struct ip_hdr  *ip_hdr;
	struct udp_hdr *udp_hdr;

    /* UDP Header */
	ptr -= sizeof(*udp_hdr);
	len += sizeof(*udp_hdr);

	udp_hdr = (struct udp_hdr *) ptr;
	udp_hdr->udp_src = event ? htons(EVENT_PORT) : htons(GENERAL_PORT);
	udp_hdr->udp_dst = event ? htons(EVENT_PORT) : htons(GENERAL_PORT);
	udp_hdr->udp_len = htons(len);

	/* IP Header */
	ptr -= sizeof(*ip_hdr);
	ip_hdr = (struct ip_hdr *) ptr;

	ip_hdr->ip_src.s_addr = raw->src_ip.s_addr;

	if (msg->header.flagField[0] & UNICAST) {
	    ip_hdr->ip_dst.s_addr = msg->address.peer_ip.sin.sin_addr.s_addr;
	    ip_hdr->ip_ttl = raw->ttl;
	} else {
	    ip_hdr->ip_dst.s_addr = mcast_addr[MC_PRIMARY].s_addr;
	    ip_hdr->ip_ttl = raw->ttl;
	}

	ip_hdr->ip_p = 17;

	udp_hdr->udp_check = 0;
	udp_hdr->udp_check = udp_checksum(ip_hdr, ptr + sizeof(*ip_hdr), len);

	len += sizeof(*ip_hdr);
	ip_hdr->ip_v = 4;
	ip_hdr->ip_hl = 5;
	ip_hdr->ip_tos = event ? raw->dscp_event : raw->dscp_general;
	ip_hdr->ip_tos <<= 2;
	ip_hdr->ip_len = htons(len);
	ip_hdr->ip_id = htons(1);
	ip_hdr->ip_off = htons(0x4000);
	ip_hdr->ip_check = 0;
	ip_hdr->ip_check = raw_udp_common_checksum(ptr, ip_hdr->ip_hl * 4);


	if (!addr)
		addr = peer ? &raw->p2p_addr : &raw->ptp_addr;


	if (raw->vlan != 0 && raw->use_vlan) {
		/* Ethernet VLAN Header */
		ptr -= sizeof(*vlan_hdr);
		len += sizeof(*vlan_hdr);
		vlan_hdr = (struct vlan_hdr *) ptr;

		addr_to_mac(&vlan_hdr->dst, addr);
		addr_to_mac(&vlan_hdr->src, &raw->src_addr);
		vlan_hdr->tpid = htons(ETH_P_8021Q);
		vlan_hdr->tci = htons((raw->vlan | 0xC000));
		vlan_hdr->type = htons(ETH_P_IP);
	} else {
		ptr -= sizeof(*eth_hdr);
		len += sizeof(*eth_hdr);
		eth_hdr = (struct eth_hdr *) ptr;

		addr_to_mac(&eth_hdr->dst, addr);
		addr_to_mac(&eth_hdr->src, &raw->src_addr);
		eth_hdr->type = htons(ETH_P_IP);
	}

	cnt = send(fd, ptr, len, 0);
	if (cnt < 1) {
		pr_err("send failed: %d %m", errno);
		return cnt;
	}
	/*
	 * Get the time stamp right away.
	 */
	return event == TRANS_EVENT ? sk_receive(fd, pkt, len, NULL, hwts, MSG_ERRQUEUE) : cnt;
}

static void raw_udp_release(struct transport *t)
{
	struct raw *raw = container_of(t, struct raw, t);
	free(raw);
}

static int raw_udp_physical_addr(struct transport *t, uint8_t *addr)
{
	struct raw *raw = container_of(t, struct raw, t);
	addr_to_mac(addr, &raw->src_addr);
	return MAC_LEN;
}

static int raw_udp_protocol_addr(struct transport *t, uint8_t *addr)
{
	struct raw *raw = container_of(t, struct raw, t);
	addr_to_mac(addr, &raw->src_addr);
	return MAC_LEN;
}

struct transport *raw_udp_transport_create(void)
{
	struct raw *raw;
	raw = calloc(1, sizeof(*raw));
	if (!raw)
		return NULL;
	raw->t.close   = raw_udp_close;
	raw->t.open    = raw_udp_open;
	raw->t.recv    = raw_udp_recv;
	raw->t.send    = raw_udp_send;
	raw->t.release = raw_udp_release;
	raw->t.physical_addr = raw_udp_physical_addr;
	raw->t.protocol_addr = raw_udp_protocol_addr;
	return &raw->t;
}
