/**
 * @file ts_monitor.c
 * Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * @note Copyright (C) 2021 Saravanan Arun <sarun@nvidia.com>
 * @note SPDX-License-Identifier: GPL-2.0+
 */
#include <stdbool.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>

#include "ts_monitor.h"
#include "print.h"

key_t tsdata_shmkey = 0xFEE1;
key_t minvio_shmkey = 0xFEE2;
key_t maxvio_shmkey = 0xFEE3;
key_t mpdvio_shmkey = 0xFEE4;
key_t tsent_shmkey = 0xFEE5;

struct ts_monitor {
	int tsdata_shmid;
	int minvio_shmid;
	int maxvio_shmid;
	int mpdvio_shmid;
	int tsent_shmid;
	int active;
	struct tsproc *tsp;
	struct time_stamp_data  *tsdata;
	struct time_stamp_entry *offset_min_vio_log;
	struct time_stamp_entry *offset_max_vio_log;
	struct time_stamp_entry *mpd_vio_log;
	struct time_stamp_entry *entry;
};

static bool ts_monitor_active(struct ts_monitor *tsm)
{
	return tsm->active ? true : false;
}

struct ts_monitor *ts_monitor_create(struct config *config, struct tsproc *tsp)
{
	struct ts_monitor *tsm;

	tsm = calloc(1, sizeof(*tsm));
	if (!tsm) {
		return NULL;
	}
	memset(tsm, 0, sizeof(*tsm));
	tsm->active = 1;
	if (!tsm->active) {
		/* Return an inactive monitor. */
		return tsm;
	}
	tsm->tsp = tsp;


	if ((tsm->tsdata_shmid = shmget(tsdata_shmkey,
									sizeof(struct time_stamp_data),
									IPC_CREAT | 0666)) < 0) {
		pr_err("shmget time_stamp_data");
		free(tsm);
		return NULL;
	}

	if ((tsm->tsdata = shmat(tsm->tsdata_shmid, NULL, 0)) == (void *) -1) {
		pr_err("shmat time_stamp_data");
		free(tsm);
		return NULL;
	}
	tsm->tsdata->wr_idx = 0;
	tsm->tsdata->offset_min_vio_log_index = 0;
	tsm->tsdata->offset_max_vio_log_index = 0;
	tsm->tsdata->mpd_vio_log_index = 0;
	tsm->tsdata->offset_min_threshold = config_get_int(config, NULL,
			               "offset_from_master_min_threshold");
	tsm->tsdata->offset_max_threshold = config_get_int(config, NULL,
			               "offset_from_master_max_threshold");
	tsm->tsdata->mpd_threshold = config_get_int(config, NULL,
			               "mean_path_delay_threshold");
	tsm->tsdata->num_ts = config_get_int(config, NULL, "tsmonitor_num_ts");
	tsm->tsdata->num_log_sets = config_get_int(config, NULL,
					"tsmonitor_num_log_sets");
	tsm->tsdata->num_log_entries = config_get_int(config, NULL,
					"tsmonitor_num_log_entries");
	tsm->tsdata->log_wait_secs = config_get_int(config, NULL,
					"tsmonitor_log_wait_seconds");

	if ((tsm->minvio_shmid = shmget(minvio_shmkey, tsm->tsdata->num_log_sets *
									tsm->tsdata->num_log_entries *
								   	sizeof(struct time_stamp_entry),
									IPC_CREAT | 0666)) < 0) {
		pr_err("shmget min_vio_log");
		free(tsm);
		return NULL;
	}

	if ((tsm->offset_min_vio_log = shmat(tsm->minvio_shmid, NULL, 0))
					== (void *) -1) {
		pr_err("shmat min_vio_log");
		free(tsm);
		return NULL;
	}

	if ((tsm->maxvio_shmid = shmget(maxvio_shmkey, tsm->tsdata->num_log_sets *
                                    tsm->tsdata->num_log_entries *
									sizeof(struct time_stamp_entry),
									IPC_CREAT | 0666)) < 0) {
		pr_err("shmget max_vio_log");
		free(tsm);
		return NULL;
	}

	if ((tsm->offset_max_vio_log = shmat(tsm->maxvio_shmid, NULL, 0))
					== (void *) -1) {
		pr_err("shmat max_vio_log");
		free(tsm);
		return NULL;
	}

	if ((tsm->mpdvio_shmid = shmget(mpdvio_shmkey, tsm->tsdata->num_log_sets *
                                    tsm->tsdata->num_log_entries *
									sizeof(struct time_stamp_entry),
									IPC_CREAT | 0666)) < 0) {
		pr_err("shmget mpd_vio_log");
		free(tsm);
		return NULL;
	}

	if ((tsm->mpd_vio_log = shmat(tsm->mpdvio_shmid, NULL, 0))
								== (void *) -1) {
		pr_err("shmat mpd_vio_log");
		free(tsm);
		return NULL;
	}

	if ((tsm->tsent_shmid = shmget(tsent_shmkey,
									tsm->tsdata->num_ts *
									sizeof(struct time_stamp_entry),
									IPC_CREAT | 0666)) < 0) {
		pr_err("shmget ts_entry");
		free(tsm);
		return NULL;
	}

	if ((tsm->entry = shmat(tsm->tsent_shmid, NULL, 0))
								== (void *) -1) {
		pr_err("shmat ts entry");
		free(tsm);
		return NULL;
	}

	return tsm;
}

void ts_monitor_destroy(struct ts_monitor *tsm)
{
	if (shmctl(tsm->tsdata_shmid, IPC_RMID, NULL) == -1)
		pr_err("shmctl tsdata");
	if (shmctl(tsm->minvio_shmid, IPC_RMID, NULL) == -1)
		pr_err("shmctl minvio");
	if (shmctl(tsm->maxvio_shmid, IPC_RMID, NULL) == -1)
		pr_err("shmctl maxvio");
	if (shmctl(tsm->mpdvio_shmid, IPC_RMID, NULL) == -1)
		pr_err("shmctl mpdvio");
	if (shmctl(tsm->tsent_shmid, IPC_RMID, NULL) == -1)
		pr_err("shmctl ts entry");
	free(tsm);
}

void ts_update_log(struct ts_monitor *tsm, int type, int32_t idx)
{
	int i, tsindex, log_set_index, log_index;
	int rolled_over = 0;
    struct time_stamp_data *tsdata = (struct time_stamp_data *)tsm->tsdata;

	if (type == 0) {
		if (tsdata->offset_min_vio_log_index < tsdata->num_log_sets) {
			log_set_index = tsdata->offset_min_vio_log_index;
			tsdata->offset_min_vio_log_index++;
		} else {
			return;
		}
	} else if (type == 1) {
		if (tsdata->offset_min_vio_log_index < tsdata->num_log_sets) {
			log_set_index = tsdata->offset_max_vio_log_index;
			tsdata->offset_max_vio_log_index++;
		} else {
			return;
		}
	} else if (type == 2) {
		if (tsdata->mpd_vio_log_index < tsdata->num_log_sets) {
			log_set_index = tsdata->mpd_vio_log_index;
			tsdata->mpd_vio_log_index++;
		} else {
			return;
		}
	} else {
		return;
	}

	log_index = ((log_set_index + 1) * tsdata->num_log_entries) - 1;
	tsindex = idx;
	i = tsdata->num_log_entries - 1;
	while ((i >= 0) && (log_index >= 0)) {
		if (type == 0)
			memcpy(&tsm->offset_min_vio_log[log_index],
				&tsm->entry[tsindex], sizeof(struct time_stamp_entry));
		else if (type == 1)
			memcpy(&tsm->offset_max_vio_log[log_index],
					&tsm->entry[tsindex], sizeof(struct time_stamp_entry));
		else if (type == 2)
			memcpy(&tsm->mpd_vio_log[log_index],
					&tsm->entry[tsindex], sizeof(struct time_stamp_entry));
		if (!rolled_over) {
  		    tsindex--;
		} else {
			tsindex++;
		}
		if (tsindex < 0) {
			tsindex = 1;
			rolled_over = 1;
		}
		i--;
		log_index--;
	}
}

void ts_monitor_update(struct ts_monitor *tsm, uint16_t seqid, tmv_t t1, tmv_t t2)
{
    struct time_stamp_data *tsdata = (struct time_stamp_data *)tsm->tsdata;
    struct timeval time, diff_time;
    tmv_t t3, t4, offset, delay;
	bool log_violation = true;
    uint16_t idx;

    if (!ts_monitor_active(tsm) || !tsdata) {
	    return;
    }

    tsproc_get_offset_delay_ts(tsm->tsp, &t3, &t4, &offset, &delay, &time);
    idx = tsdata->wr_idx;
    tsm->entry[idx].sync_seqid = seqid;
    tsm->entry[idx].t1 = t1;
    tsm->entry[idx].t2 = t2;
    tsm->entry[idx].t3 = t3;
    tsm->entry[idx].t4 = t4;
    tsm->entry[idx].master_offset = offset;
    tsm->entry[idx].path_delay = delay;
    memcpy(&tsm->entry[idx].sync_time, &time, sizeof(struct timeval));

    if ((offset.ns < 0) && (offset.ns < tsm->tsdata->offset_min_threshold)) {
		if (tsm->tsdata->log_wait_secs > 0) {
			timersub(&time, &tsdata->last_offset_min_violation, &diff_time);
			if (diff_time.tv_sec < tsm->tsdata->log_wait_secs) {
					log_violation = false;
			}
		}
		if (log_violation) {
			tsdata->offset_min_violations++;
			memcpy(&tsdata->last_offset_min_violation, &time, sizeof(struct timeval));
			ts_update_log(tsm, 0, idx);
		}
	}
	if ((offset.ns > 0) && (offset.ns > tsm->tsdata->offset_max_threshold)) {
		if (tsm->tsdata->log_wait_secs > 0) {
			timersub(&time, &tsdata->last_offset_max_violation, &diff_time);
			if (diff_time.tv_sec < tsm->tsdata->log_wait_secs) {
				log_violation = false;
			}
		}
		if (log_violation) {
			tsdata->offset_max_violations++;
			memcpy(&tsdata->last_offset_max_violation, &time, sizeof(struct timeval));
			ts_update_log(tsm, 1, idx);
		}
	}
	if (delay.ns > tsm->tsdata->mpd_threshold) {
		if (tsm->tsdata->log_wait_secs > 0) {
			timersub(&time, &tsdata->last_mpd_violation, &diff_time);
			if (diff_time.tv_sec < tsm->tsdata->log_wait_secs) {
					log_violation = false;
			}
		}
		if (log_violation) {
			tsdata->mpd_violations++;
			memcpy(&tsdata->last_mpd_violation, &time, sizeof(struct timeval));
			ts_update_log(tsm, 2, idx);
		}
	}
    if (idx++ == tsdata->num_ts) {
		idx = 0;
	}
    tsdata->wr_idx = idx;
}

