#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include "ts_monitor.h"

#define TSMON_MAX_BUF_SIZE 8192

#define TIMESTAMP_TO_NS(x) \
	((((uint64_t)x.seconds_msb) << 32 | x.seconds_lsb) * 1000000000ULL) + x.nanoseconds

static int running = 1;

key_t tsdata_shmkey = 0xFEE1;
key_t minvio_shmkey = 0xFEE2;
key_t maxvio_shmkey = 0xFEE3;
key_t mpdvio_shmkey = 0xFEE4;
key_t tsent_shmkey = 0xFEE5;

char outbuf[TSMON_MAX_BUF_SIZE];

static void handle_int_quit_term(int s)
{
    running = 0;
}

int handle_term_signals(void)
{
    if (SIG_ERR == signal(SIGINT, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGINT\n");
            return -1;
    }
    if (SIG_ERR == signal(SIGQUIT, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGQUIT\n");
            return -1;
    }
    if (SIG_ERR == signal(SIGTERM, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGTERM\n");
            return -1;
    }
    if (SIG_ERR == signal(SIGHUP, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGHUP\n");
            return -1;
    }
    return 0;
}

void dt_time_usec(struct timeval *tv, char *buf)
{
    time_t ntime;
    struct tm ntm;
    char nbuf[32];
    int sz;

	ntime = tv->tv_sec;
	localtime_r(&ntime, &ntm);
	sz = strftime(nbuf, 30, "%Y-%m-%dT%H:%M:%S", &ntm);
    sprintf(buf,"%*s.%06dZ", sz, nbuf, (int)tv->tv_usec);
}

int main(int argc, char *argv[])
{
    struct time_stamp_entry *offset_min_vio_log;
    struct time_stamp_entry *offset_max_vio_log;
    struct time_stamp_entry *mpd_vio_log;
    struct time_stamp_entry *entry;
    struct time_stamp_data *tsdata;
    struct option opts;
    int tsdata_shmid;
    int minvio_shmid;
    int maxvio_shmid;
    int mpdvio_shmid;
    int tsent_shmid;
    char buf[32], buf1[32], buf2[32];
    uint16_t idx, logidx;
    int index, i;
    int opt, opt1, wrlen = 0, len = 0;

    handle_term_signals();

    opt= getopt_long(argc, argv, "oxvmnpcg", &opts, &index);
    opt1= getopt_long(argc, argv, "j", &opts, &index);
    if (opt1 < 0) {
        switch (opt) {
        case 'o':
            fprintf(stdout, "Time (UTC)               Seq ID Offset (ns)           MPD (ns)\n");
            fprintf(stdout, "------------------------ ------ -------------------- --------------------\n");
            break;
        case 'x':
            fprintf(stdout, "Time (UTC)               Seq ID Offset (ns)          MPD (ns)             T1 (ns)              T2 (ns)              T3 (ns)              T4 (ns)\n");
            fprintf(stdout, "------------------------ ------ -------------------- -------------------- -------------------- -------------------- -------------------- --------------------\n");
            break;
        case 'v':
            break;
        case 'm':
        case 'n':
        case 'p':
        fprintf(stdout, "Time (UTC)               Seq ID Offset (ns)          MPD (ns)             T1 (ns)              T2 (ns)              T3 (ns)              T4 (ns)\n");
        fprintf(stdout, "------------------------ ------ -------------------- -------------------- -------------------- -------------------- -------------------- --------------------\n\n");
            break;
        case 'c':
        case 'g':
            break;
        default:
            fprintf(stderr,
                "\nusage: tsmonitor [options][-j]\n\n"
                " -o    Master Offset information\n"
                " -x    Extended Master Offset information\n"
                " -v    Threshold violations\n"
	            " -m    Max Threshold violations log \n"\
	            " -n    Min Threshold violations log \n"\
	            " -p    MPD Threshold violations log \n"\
	            " -c    Clear Threshold violation counters \n"
    	        " -g    Clear Threshold violations log \n"
    	        " -j    JSON option\n\n");
            return -1;
        }
    }
    if ((tsdata_shmid = shmget(tsdata_shmkey, sizeof(struct time_stamp_data), 0666)) < 0) {
        perror("shmget time_stamp_data");
        return -1;
    }

    if ((tsdata = shmat(tsdata_shmid, NULL, 0)) == (void *) -1) {
        perror("shmat time_stamp_data");
        return -1;
    }

    if ((minvio_shmid = shmget(minvio_shmkey, tsdata->num_log_sets *
                                    tsdata->num_log_entries *
                                    sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget min_vio_log");
        return -1;
    }

    if ((offset_min_vio_log = shmat(minvio_shmid, NULL, 0))
                    == (void *) -1) {
        printf("shmat min_vio_log");
        return -1;
    }

    if ((maxvio_shmid = shmget(maxvio_shmkey, tsdata->num_log_sets *
                                    tsdata->num_log_entries *
                                    sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget max_vio_log");
        return -1;
    }

    if ((offset_max_vio_log = shmat(maxvio_shmid, NULL, 0))
                    == (void *) -1) {
        printf("shmat max_vio_log");
        return -1;
    }

    if ((mpdvio_shmid = shmget(mpdvio_shmkey, tsdata->num_log_sets *
                                    tsdata->num_log_entries *
                                    sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget mpd_vio_log");
        return -1;
    }

    if ((mpd_vio_log = shmat(mpdvio_shmid, NULL, 0))
                                == (void *) -1) {
        printf("shmat mpd_vio_log");
        return -1;
    }

    if ((tsent_shmid = shmget(tsent_shmkey,
                            tsdata->num_ts * sizeof(struct time_stamp_entry),
                            IPC_CREAT | 0666)) < 0) {
        printf("shmget ts_entry");
        return -1;
    }

    if ((entry = shmat(tsent_shmid, NULL, 0))
                                == (void *) -1) {
        printf("shmat ts entry");
        return -1;
    }

    if (opt == 'v') {
        if (opt1 == 'j') {
            if (tsdata->offset_min_violations) {
                dt_time_usec(&tsdata->last_offset_min_violation, buf);
    		}
            if (tsdata->offset_max_violations) {
                dt_time_usec(&tsdata->last_offset_max_violation, buf1);
            }
            if (tsdata->mpd_violations) {
                dt_time_usec(&tsdata->last_mpd_violation, buf2);
            }
            wrlen = sprintf(outbuf, 
				"min-offset-count=%d,last-min-offset=%s,max-offset-count=%d,last-max-offset=%s,path-delay-count=%d,last-path-delay=%s\n",
                tsdata->offset_min_violations, 
                tsdata->offset_min_violations ? buf : " ",
                tsdata->offset_max_violations,
                tsdata->offset_max_violations ? buf1 : " ",
                tsdata->mpd_violations,
                tsdata->mpd_violations ? buf2 : " ");
            printf("%s", outbuf);
        } else {
		    fprintf(stdout,
	            "Maximum number of Time Stamp Entries       : %d\n"
    	        "Maximum number of Violation log sets       : %d\n"
    	        "Maximum number of Violation entries per set: %d\n"
    	        "Violation log wait interval (sec)          : %d\n\n"
        		"Minimum Offset Threshold                   : %d\n"
        		"Maximum Offset Threshold                   : %d\n"
        		"Mean Path Delay Threshold                  : %d\n\n",
        		tsdata->num_ts, tsdata->num_log_sets, 
                tsdata->num_log_entries,
			    tsdata->log_wait_secs, tsdata->offset_min_threshold,
			    tsdata->offset_max_threshold, tsdata->mpd_threshold);

            if (tsdata->offset_min_violations) {
                dt_time_usec(&tsdata->last_offset_min_violation, buf);
    		}
    		fprintf(stdout,
    	        "Minimum Offset Violations : %6d Last Occurrence: %s\n",
        		tsdata->offset_min_violations,
        		tsdata->offset_min_violations ? buf : " ");

            if (tsdata->offset_max_violations) {
                dt_time_usec(&tsdata->last_offset_max_violation, buf);
            }

		    fprintf(stdout,
                "Maximum Offset Violations : %6d Last Occurrence: %s\n",
                tsdata->offset_max_violations,
                tsdata->offset_max_violations ? buf : " ");

            if (tsdata->mpd_violations) {
                dt_time_usec(&tsdata->last_mpd_violation, buf);
            }
            fprintf(stdout,
                "Mean Path Delay Violations: %6d Last Occurrence: %s\n",
                tsdata->mpd_violations,
                tsdata->mpd_violations ? buf : " ");

            return 0;
        }
    } else if ((opt == 'n') && (tsdata->offset_min_vio_log_index)) {
		for (idx = 0, logidx = 0; (idx < tsdata->offset_min_vio_log_index) &&
				(idx < tsdata->num_log_sets); ++idx) {
            if (opt1 != 'j') {
  			    fprintf(stdout, 
                    "\nMinimum Offset violations log [%d]:\n\n", idx);
            }
			for (i = 0; i < tsdata->num_log_entries; ++i, ++logidx) {
				if (!offset_min_vio_log[logidx].sync_seqid &&
					!offset_min_vio_log[logidx].master_offset.ns &&
					!offset_min_vio_log[logidx].path_delay.ns) {
					continue;
				}
                dt_time_usec(&offset_min_vio_log[logidx].sync_time, buf);
                if (opt1 == 'j') {
                    wrlen = sprintf(outbuf + len,
  			                    "%d-%d@log-time=%s,offset-from-master=%"PRId64",mean-path-delay=%"PRId64",t1=%"PRId64",t2=%"PRId64",t3=%"PRId64",t4=%"PRId64"\n",
						offset_min_vio_log[logidx].sync_seqid, idx, buf,
						offset_min_vio_log[logidx].master_offset.ns,
						offset_min_vio_log[logidx].path_delay.ns,
						offset_min_vio_log[logidx].t1.ns,
						offset_min_vio_log[logidx].t2.ns,
						offset_min_vio_log[logidx].t3.ns,
						offset_min_vio_log[logidx].t4.ns);
                        len += wrlen;
                        if ((TSMON_MAX_BUF_SIZE - len) < 204)
                            break;

                } else {
   					fprintf(stdout,
				        "%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n" ,
						buf, offset_min_vio_log[logidx].sync_seqid,
						offset_min_vio_log[logidx].master_offset.ns,
						offset_min_vio_log[logidx].path_delay.ns,
						offset_min_vio_log[logidx].t1.ns,
						offset_min_vio_log[logidx].t2.ns,
						offset_min_vio_log[logidx].t3.ns,
						offset_min_vio_log[logidx].t4.ns);
                }
		  	}
	    }
        if (opt1 == 'j')
            printf("%s", outbuf);
    }
    if ((opt == 'm') && (tsdata->offset_max_vio_log_index)) {
	    for (idx = 0, logidx = 0;
                (idx < tsdata->offset_max_vio_log_index) &&
			    (idx < tsdata->num_log_sets); ++idx) {
            if (opt1 != 'j') {
  	            fprintf(stdout,
                    "\nMaximum Offset violations log [%d]:\n\n", idx);
            }
			for (i = 0; i < tsdata->num_log_entries; ++i, ++logidx) {
				if (!offset_max_vio_log[logidx].sync_seqid &&
					!offset_max_vio_log[logidx].master_offset.ns &&
					!offset_max_vio_log[logidx].path_delay.ns) {
					continue;
				}
                dt_time_usec(&offset_max_vio_log[logidx].sync_time, buf);
                if (opt1 == 'j') {
                    wrlen = sprintf(outbuf + len,
  			                    "%d-%d@log-time=%s,offset-from-master=%"PRId64",mean-path-delay=%"PRId64",t1=%"PRId64",t2=%"PRId64",t3=%"PRId64",t4=%"PRId64"\n",
						offset_min_vio_log[logidx].sync_seqid, idx, buf,
						offset_min_vio_log[logidx].master_offset.ns,
						offset_min_vio_log[logidx].path_delay.ns,
						offset_min_vio_log[logidx].t1.ns,
						offset_min_vio_log[logidx].t2.ns,
						offset_min_vio_log[logidx].t3.ns,
						offset_min_vio_log[logidx].t4.ns);
                        len += wrlen;
                        if ((TSMON_MAX_BUF_SIZE - len) < 204)
                            break;
                } else {
					fprintf(stdout,
				"%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n" ,
						buf, offset_max_vio_log[logidx].sync_seqid,
						offset_max_vio_log[logidx].master_offset.ns,
						offset_max_vio_log[logidx].path_delay.ns,
						offset_max_vio_log[logidx].t1.ns,
						offset_max_vio_log[logidx].t2.ns,
						offset_max_vio_log[logidx].t3.ns,
						offset_max_vio_log[logidx].t4.ns);
				}
		    }
	    }
        if (opt1 == 'j')
            printf("%s", outbuf);
    }
	if ((opt == 'p') && (tsdata->mpd_vio_log_index)) {
	    for (idx = 0, logidx = 0; (idx < tsdata->mpd_vio_log_index) &&
			  (idx < tsdata->num_log_sets); ++idx) {
            if (opt1 != 'j') {
  	            fprintf(stdout,
                    "\nMean Path Delay violations log [%d]:\n\n", idx);
            }
			for (i = 0; i < tsdata->num_log_entries; ++i, ++logidx) {
				if (!mpd_vio_log[logidx].sync_seqid &&
					!mpd_vio_log[logidx].master_offset.ns &&
					!mpd_vio_log[logidx].path_delay.ns) {
					continue;
				}
                dt_time_usec(&mpd_vio_log[logidx].sync_time, buf);

                if (opt1 == 'j') {
                    wrlen = sprintf(outbuf + len,
  			                    "%d-%d@log-time=%s,offset-from-master=%"PRId64",mean-path-delay=%"PRId64",t1=%"PRId64",t2=%"PRId64",t3=%"PRId64",t4=%"PRId64"\n",
						offset_min_vio_log[logidx].sync_seqid, idx, buf,
						offset_min_vio_log[logidx].master_offset.ns,
						offset_min_vio_log[logidx].path_delay.ns,
						offset_min_vio_log[logidx].t1.ns,
						offset_min_vio_log[logidx].t2.ns,
						offset_min_vio_log[logidx].t3.ns,
						offset_min_vio_log[logidx].t4.ns);
                        len += wrlen;
                        if ((TSMON_MAX_BUF_SIZE - len) < 204)
                            break;
                } else {
					fprintf(stdout,
				"%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n" ,
						buf, mpd_vio_log[logidx].sync_seqid,
						mpd_vio_log[logidx].master_offset.ns,
						mpd_vio_log[logidx].path_delay.ns,
						mpd_vio_log[logidx].t1.ns,
						mpd_vio_log[logidx].t2.ns,
						mpd_vio_log[logidx].t3.ns,
						mpd_vio_log[logidx].t4.ns);
				}
		    }
	    }
        if (opt1 == 'j')
            printf("%s", outbuf);
    } else if (opt == 'c') {
	    memset(&tsdata->last_offset_min_violation, 0, sizeof(struct timeval));
	    memset(&tsdata->last_offset_max_violation, 0, sizeof(struct timeval));
	    memset(&tsdata->last_mpd_violation, 0, sizeof(struct timeval));

	    tsdata->offset_min_violations = 0;
	    tsdata->offset_max_violations = 0;
	    tsdata->mpd_violations = 0;

	    return 0;
    } else if (opt == 'g') {
	    tsdata->offset_min_vio_log_index = 0;
	    tsdata->offset_max_vio_log_index = 0;
	    tsdata->mpd_vio_log_index = 0;

	    return 0;
    } else if (opt == 'o') {
        while (idx++ < tsdata->num_ts) {
    		if (!entry[idx].sync_seqid &&
    			!entry[idx].master_offset.ns &&
    			!entry[idx].path_delay.ns) {
    			break;
    		}

            dt_time_usec(&entry[idx].sync_time, buf);
			if (opt1 == 'j') {
	            wrlen = sprintf(outbuf + len,
				"%d@log-time=%s,offset-from-master=%"PRId64",mean-path-delay=%"PRId64"\n",
                entry[idx].sync_seqid, buf, entry[idx].master_offset.ns,
                entry[idx].path_delay.ns);
                len += wrlen;
                if ((TSMON_MAX_BUF_SIZE - len) < 100)
                    break;
			} else {
				fprintf(stdout, "%24s %6d %20" PRId64" %20" PRId64 "\n",
                buf, entry[idx].sync_seqid, entry[idx].master_offset.ns,
                entry[idx].path_delay.ns);
			}
        }
        if (opt1 == 'j')
            printf("%s", outbuf);
    } else if (opt == 'x') {
        while (idx++ < tsdata->num_ts) {
    		if (!entry[idx].sync_seqid &&
    			!entry[idx].master_offset.ns &&
    			!entry[idx].path_delay.ns) {
    			break;
    		}
            dt_time_usec(&entry[idx].sync_time, buf);
			if (opt1 == 'j') {
	            wrlen = sprintf(outbuf + len,
				"%d@log-time=%s,offset-from-master=%"PRId64",mean-path-delay=%"PRId64",t1=%"PRId64",t2=%"PRId64",t3=%"PRId64",t4=%"PRId64"\n",
				entry[idx].sync_seqid, buf, entry[idx].master_offset.ns,
				entry[idx].path_delay.ns, entry[idx].t1.ns,
				entry[idx].t2.ns, entry[idx].t3.ns,
				entry[idx].t4.ns);
                len += wrlen;
                if ((TSMON_MAX_BUF_SIZE - len) < 204)
                    break;
			} else {
				fprintf(stdout, "%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n",
				buf, entry[idx].sync_seqid, entry[idx].master_offset.ns,
				entry[idx].path_delay.ns, entry[idx].t1.ns,
				entry[idx].t2.ns, entry[idx].t3.ns,
				entry[idx].t4.ns);
			}
		}
        if (opt1 == 'j')
            printf("%s", outbuf);
    }
}
